@@ -2,6 +2,7 @@ import nanoscipy.util as nsu
 import numpy as np
 import sympy as sp
 import scipy.constants as spc
+from itertools import chain
 
 
 def basic_operations(operator, fir_int, sec_int=None):
@@ -20,6 +21,16 @@ def basic_operations(operator, fir_int, sec_int=None):
     Returns
         Product of the operation.
     """
+
+    # check if any of the variables are of type nan
+    if 'n' in (fir_int, sec_int):
+        # if true, output nan result
+        return np.float64('nan')
+    elif isinstance(fir_int, str):
+        raise ValueError(f'No value found for \'{fir_int}\'.')
+    elif isinstance(sec_int, str):
+        raise ValueError(f'No value found for \'{sec_int}\'.')
+
     if operator == '+':
         opr_res = fir_int + sec_int
     elif operator in ('-', '+-'):
@@ -184,16 +195,22 @@ def number_parser(math_string):
                                                                                                         ip4)]
             elif (ip1_val, ip2_val) == ('N', 'A'):
                 temp_string = [str(spc.N_A) if k == i else j for k, j in temp_index_chain if k not in (ip1, ip2)]
-            elif (ip1_val, ip2_val) == ('k', 'B'):
-                temp_string = [str(spc.k) if k == i else j for k, j in temp_index_chain if k not in (ip1, ip2)]
-            elif (ip1_val, ip2_val) == ('R', 'c'):
-                temp_string = [str(spc.R) if k == i else j for k, j in temp_index_chain if k not in (ip1, ip2)]
+            elif (ip1_val, ip2_val) == ('m', 'e'):
+                temp_string = [str(spc.electron_mass) if k == i else j for k, j in temp_index_chain if k not in
+                               (ip1, ip2)]
+            elif (ip1_val, ip2_val) == ('m', 'p'):
+                temp_string = [str(spc.proton_mass) if k == i else j for k, j in temp_index_chain if k not in
+                               (ip1, ip2)]
             elif ip1_val == 'c':
                 temp_string = [str(spc.c) if k == i else j for k, j in temp_index_chain if k != ip1]
             elif ip1_val == 'h':
                 temp_string = [str(spc.h) if k == i else j for k, j in temp_index_chain if k != ip1]
             elif ip1_val == 'e':
                 temp_string = [str(spc.e) if k == i else j for k, j in temp_index_chain if k != ip1]
+            elif ip1_val == 'R':
+                temp_string = [str(spc.R) if k == i else j for k, j in temp_index_chain if k != ip1]
+            elif ip1_val == 'k':
+                temp_string = [str(spc.k) if k == i else j for k, j in temp_index_chain if k != ip1]
             else:
                 raise ValueError('Constant is not defined in parser.')
             temp_index_chain = nsu.indexer(temp_string)
@@ -280,49 +297,63 @@ def ordered_parser(math_string, steps=False):
     return o4_opr_string[0]
 
 
-def product_parser(math_string):
+def product_parser(string, items):
     """
-    Parser that detects implicit multiplication and adds appropriate operators in those positions.
+        Parser that detects implicit multiplication and adds multiplication operator in those positions. Note
+        importantly that this script works by splitting the string around every exclusion (the largest exclusion first -
+        once an exclusion has been split, it cannot be further split), and then adding an operator in every gap in the
+        new list, if the last value of the current string is not a mathematical symbol, and the first value of the next
+        string is not a mathematical symbol.
 
-    Parameters
-        math_string : string
-            The mathematical string to search for implicit multiplication.
+        Parameters
+            string : str
+                The mathematical string to search for implicit multiplication.
+            items : str, tuple
+                Items that implicit multiplication will be done around if appropriate.
 
-    Returns
-        Updated list with the implicit multiplication as explicit.
-    """
-    temp_decom_string = [e for e in math_string]  # temp string to be updated
-    i = 0  # initial iteration
-    while i < len(temp_decom_string):
-        i0_val = temp_decom_string[i]  # define current iteration value
-        ip1, ip2, ip3 = i + 1, i + 2, i + 3  # define consecutive iterations
-        ip1_val = ip2_val = ip3_val = None  # provide standard values for those iterations
-        try:  # try to find values for those iterations
-            ip1_val = temp_decom_string[ip1]
-            ip2_val = temp_decom_string[ip2]
-            ip3_val = temp_decom_string[ip3]
-        except IndexError:  # if index is surpassed, pass clause at position
-            pass
+        Returns
+            Updated string with the implicit multiplication as explicit.
+        """
 
-        # determine in which positions to add a '*' and add it
-        if (i0_val, ip1_val) == (')', '(') or (isinstance(nsu.string_to_float(i0_val), float) and ip1_val in
-                                               (*[i for i in nsu.alphabetSequence if i != 'e'],
-                                                *nsu.alphabetSequenceCap, '(')) or \
-                (i0_val in (*[i for i in nsu.alphabetSequence if i != 'e'], *nsu.alphabetSequenceCap, ')') and
-                 isinstance(nsu.string_to_float(ip1_val), float)) or \
-                (isinstance(nsu.string_to_float(i0_val), float) and (ip1_val, ip2_val, ip3_val) == ('e', 'x', 'p')) or \
-                (i0_val == ')' and ip1_val in (*nsu.alphabetSequence, *nsu.alphabetSequenceCap)) or \
-                ((i0_val, ip1_val) == (')', '_')) or \
-                (isinstance(nsu.string_to_float(i0_val), float) and ip1_val == '_') or \
-                (i0_val in (*nsu.alphabetSequence, *nsu.alphabetSequenceCap) and ip1_val == '_'):
-            temp_decom_string = temp_decom_string[: ip1] + ['*'] + temp_decom_string[ip1:]
-        elif (i0_val, ip1_val, ip2_val) == ('p', 'i', '('):
-            temp_decom_string = temp_decom_string[: ip2] + ['*'] + temp_decom_string[ip2:]
-        i += 1  # update iterator
-    return temp_decom_string
+    # split the given string around the given items, making sure that the largest items are iterated through first
+    sorted_items = nsu.string_sorter(items + (')',), reverse=True)
+    split_list = nsu.multi_split(string, sorted_items, False)
+
+    # remove any blank fields if present
+    no_blanks_itr_str = [i for i in split_list if i != '']
+
+    # add implicit multiplication for every gap in list, if the adjacent elements are not mathematical symbols
+    i0 = 0
+    temp_list = no_blanks_itr_str
+    while i0 < len(temp_list):
+
+        # define initial values
+        ip1 = i0 + 1
+        i0_val = temp_list[i0]
+
+        # try to define real values, and break if not possible (as there is then no more elements)
+        try:
+            ip1_val = temp_list[ip1]
+        except IndexError:
+            break
+
+        # define mathematical symbols
+        math_symbols_gen = ('-', '+', '/', '*', '^')
+        math_symbols_i0 = ('(',) + math_symbols_gen
+        math_symbols_ip1 = ('!', ')') + math_symbols_gen
+
+        # if i0_val and ip1_val does not have conflicting mathematical symbols, add '*' to i0_val and update i0
+        if i0_val[-1] not in math_symbols_i0 and ip1_val[0] not in math_symbols_ip1:
+            temp_list[i0] = i0_val + '*'
+        i0 += 1
+
+    # revert list to a string
+    result_string = nsu.list_to_string(temp_list)
+
+    return result_string
 
 
-def parser(math_string, steps=False, cprint=True):
+def parser(math_string, steps=False, cprint='num', **kwargs):
     """
     Takes care of the additional rules and conventions of mathematical operations. Handles parentheses along with
     operators that require parentheses, such as trigonometric functions (sin, cos, tan, ...) and log, exp, etc.
@@ -332,14 +363,26 @@ def parser(math_string, steps=False, cprint=True):
             The mathematical string to parse through the interpreter.
         steps : bool, optional
             If True, displays balances from the script, whilst performing the operations. The default is False.
-        cprint : bool, optional
-            If True, prints the calculation and result. The default is True.
+        cprint : str (or False), optional
+                Determines whether the computational result should be printed in the python console. There are four
+                options: 'num' will display input string with constants replaced with values, 'sym' will display input
+                string with constants as symbols, False will disable result print. The default is 'num'.
+
+    Keyword Arguments
+        true_string : str
+            A true input string, used when cprint is set, to swap symbols in etc. or simply for a prettier cprint.
 
     Returns
         The result from the performed operations on the given mathematical string as a float.
     """
     # define temporary lists/values to be updated from while loop
-    temp_decom_string = product_parser(math_string)
+    # first define items for product_parser
+    constant_items = ('pi', '_hbar', '_NA', '_c', '_h', '_R', '_k', '_e', '_me', '_mp')
+    function_items = ('sinh(', 'cosh(', 'tanh(', 'exp(', 'sin(', 'cos(', 'tan(', 'ln(', 'rad(',
+                      'deg(', 'log(', 'sqrt(', 'arcsin(', 'arccos(', 'arctan(', 'arcsinh(', 'arccosh(', 'arctanh(')
+    collective_items = constant_items + function_items
+
+    temp_decom_string = product_parser(math_string, collective_items)
     temp_index = nsu.indexer(temp_decom_string)
     temp_bracket_idx = [[j] + [e] for j, e in temp_index if e in ('(', ')')]  # find and index open/close brackets
     if steps:
@@ -348,9 +391,17 @@ def parser(math_string, steps=False, cprint=True):
     while temp_bracket_idx:
         # if two consecutive brackets are a pair, execute operations through ordered_parser(), append the result to the
         #   given string, update it and reiterate. This ensures that the parentheses are read in the correct order
-        if temp_bracket_idx[i][1] == '(' and temp_bracket_idx[i + 1][1] == ')':
+
+        try:  # check for missing closing parenthesis
+            closing_bracket = temp_bracket_idx[i + 1][1]
+        except IndexError:
+            raise ValueError('Missing closing bracket somewhere.')
+
+        if temp_bracket_idx[i][1] == '(' and closing_bracket == ')':
             i0, i1 = temp_bracket_idx[i][0], temp_bracket_idx[i + 1][0]  # define current i'th values
-            im1, im2, im3, im4, im5, im6 = i0 - 1, i0 - 2, i0 - 3, i0 - 4, i0 - 5, i0 - 6  # define consecutive -i's
+
+            # define consecutive -i's
+            im1, im2, im3, im4, im5, im6, im7 = i0 - 1, i0 - 2, i0 - 3, i0 - 4, i0 - 5, i0 - 6, i0 - 7
             ip1 = i1 + 1  # define consecutive +i's
             bracket_excl = list(range(i0 + 1, ip1))  # define the bracket clause as an exclusion
             new_string = nsu.list_to_string(temp_decom_string[i0 + 1: i1])  # string consisting only of the clause
@@ -359,7 +410,7 @@ def parser(math_string, steps=False, cprint=True):
             # define temporary lists/values
             id_excl = []
             temp_result = pre_temp_result
-            im6_val = im5_val = im4_val = im3_val = im2_val = im1_val = ip1_val = None
+            im7_val = im6_val = im5_val = im4_val = im3_val = im2_val = im1_val = ip1_val = None
             try:  # try to define values for the surrounding iterations, otherwise pass at position
                 im1_val = temp_decom_string[im1]
                 im2_val = temp_decom_string[im2]
@@ -367,6 +418,7 @@ def parser(math_string, steps=False, cprint=True):
                 im4_val = temp_decom_string[im4]
                 im5_val = temp_decom_string[im5]
                 im6_val = temp_decom_string[im6]
+                im7_val = temp_decom_string[im7]
             except IndexError:
                 pass
             try:
@@ -376,18 +428,36 @@ def parser(math_string, steps=False, cprint=True):
 
             # if preceding iterations or upcoming operations leads to an identifier, perform special operation on the
             #   clause, respecting order. From here, define temporary result to append, along with index for exclusions
-            if (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 's', 'i', 'n') or \
-                    (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('s', 'i', 'n', '^', '-', '1'):
+            if (im7_val, im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 's', 'i', 'n', 'h'):
+                temp_result = np.arcsinh(pre_temp_result)
+                id_excl = list(range(im7, i0))
+            elif (im7_val, im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 'c', 'o', 's', 'h'):
+                temp_result = np.arccosh(pre_temp_result)
+                id_excl = list(range(im7, i0))
+            elif (im7_val, im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 't', 'a', 'n', 'h'):
+                temp_result = np.arctanh(pre_temp_result)
+                id_excl = list(range(im7, i0))
+            elif (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 's', 'i', 'n'):
                 temp_result = np.arcsin(pre_temp_result)
                 id_excl = list(range(im6, i0))
-            elif (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 'c', 'o', 's') or \
-                    (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('c', 'o', 's', '^', '-', '1'):
+            elif (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 'c', 'o', 's'):
                 temp_result = np.arccos(pre_temp_result)
                 id_excl = list(range(im6, i0))
-            elif (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 't', 'a', 'n') or \
-                    (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('t', 'a', 'n', '^', '-', '1'):
+            elif (im6_val, im5_val, im4_val, im3_val, im2_val, im1_val) == ('a', 'r', 'c', 't', 'a', 'n'):
                 temp_result = np.tan(pre_temp_result)
                 id_excl = list(range(im6, i0))
+            elif (im4_val, im3_val, im2_val, im1_val) == ('s', 'i', 'n', 'h'):
+                temp_result = np.sinh(pre_temp_result)
+                id_excl = list(range(im4, i0))
+            elif (im4_val, im3_val, im2_val, im1_val) == ('c', 'o', 's', 'h'):
+                temp_result = np.cosh(pre_temp_result)
+                id_excl = list(range(im4, i0))
+            elif (im4_val, im3_val, im2_val, im1_val) == ('t', 'a', 'n', 'h'):
+                temp_result = np.tanh(pre_temp_result)
+                id_excl = list(range(im4, i0))
+            elif (im4_val, im3_val, im2_val, im1_val) == ('s', 'q', 'r', 't'):
+                temp_result = pre_temp_result ** (1 / 2)
+                id_excl = list(range(im4, i0))
             elif (im3_val, im2_val, im1_val) == ('e', 'x', 'p'):
                 temp_result = np.exp(pre_temp_result)
                 id_excl = list(range(im3, i0))
@@ -446,11 +516,33 @@ def parser(math_string, steps=False, cprint=True):
 
     # auto-print if prompted
     if cprint:
-        pretty_string = nsu.replace(('pi', '_hbar', '_NA', '_c', '_h', '*', '_Rc', '_kB', '_e'),
-                                    ('π', 'ħ', 'Nᴀ', 'c', 'h', '⋅', 'Rᴄ', 'kᴮ', 'e'), math_string)
+        # pi-prettify if cprint has any value
         if isinstance(nsu.float_to_int(int_fixed_string / np.pi), int) and int_fixed_string != 0:
             pi_fixed_string = str(nsu.float_to_int(int_fixed_string / np.pi)) + 'π'
         else:
             pi_fixed_string = int_fixed_string
-        print(f'Result: {pretty_string} = {pi_fixed_string}')
+
+            # check whether a true input string has been given, along with added unit result
+        true_string, unit_res = math_string, ''
+        if 'true_string' in kwargs.keys():
+            true_string = kwargs.get('true_string')
+        if 'unit_res' in kwargs.keys():
+            unit_res = kwargs.get('unit_res')
+
+        # define specific set of replacement keys/values depending on cprint type
+        if cprint == 'num':
+            replacement_keys = ('pi', '_hbar', '_NA', '_c', '_h', '*', '_R', '_k', '_e', '_me', '_mp')
+            replacement_vals = ('π', 'ħ', 'Nᴀ', 'c', 'h', '⋅', 'R', 'k', 'e', 'mₑ', 'mₚ')
+        elif cprint == 'sym':
+            replacement_keys = ['_hbar', '_NA', '_c', '_h', '*', '_R', '_k', '_e', '_me', '_mp'] + \
+                               nsu.alphabetSequenceGreekLetters + nsu.alphabetSequenceGreekLettersCap
+            replacement_vals = ['ħ', 'Nᴀ', 'c', 'h', '⋅', 'R', 'k', 'e', 'mₑ', 'mₚ'] + \
+                               nsu.alphabetSequenceGreek + nsu.alphabetSequenceGreekCap
+        else:
+            raise ValueError(f'Computation print type \'{cprint}\' is not supported.')
+
+        # sort the replacements with their keys, replace them and print
+        sorted_replacements = nsu.string_sorter(replacement_keys, replacement_vals, reverse=True, otype='tuple')
+        pretty_string = nsu.replace(sorted_replacements[0], sorted_replacements[1], true_string)
+        print(f'Result: {pretty_string} = {str(pi_fixed_string) + unit_res}')
     return int_fixed_string
