@@ -21,11 +21,14 @@ float_to_int()
 
 replace()
 
+string_sorter()
+
 """
 import warnings
 import numpy as np
 import itertools
 from itertools import chain
+from operator import itemgetter
 
 standardColorsHex = ['#5B84B1FF', '#FC766AFF', '#5F4B8BFF', '#E69A8DFF', '#42EADDFF', '#CDB599FF', '#00A4CCFF',
                      '#F95700FF', '#00203FFF', '#ADEFD1FF', '#F4DF4EFF', '#949398FF', '#ED2B33FF', '#D85A7FFF',
@@ -34,6 +37,17 @@ alphabetSequence = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
                     'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
 alphabetSequenceCap = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
                        'T', 'U', 'V', 'W', 'X', 'Y', 'Z']
+alphabetSequenceGreek = ['α', 'β', 'γ', 'δ', 'ε', 'ζ', 'η', 'θ', 'ι', 'κ', 'λ', 'μ', 'ν', 'ξ', 'ο', 'π', 'ρ', 'σ', 'τ',
+                         'υ', 'φ', 'χ', 'ψ', 'ω']
+alphabetSequenceGreekCap = ['Α', 'B', 'Γ', 'Δ', 'E', 'Ζ', 'H', 'Θ', 'Ι', 'K', 'Λ', 'M', 'N', 'Ξ', '	Ο', 'Π', 'P',
+                            'Σ',
+                            'T', 'Y', 'Φ', 'X', 'Ψ', 'Ω']
+alphabetSequenceGreekLetters = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta', 'theta', 'iota', 'kappa',
+                                'lambda', 'mu', 'nu', 'xi', 'omicron', 'pi', 'rho', 'sigma', 'tau', 'upsilon', 'phi',
+                                'chi', 'psi', 'omega']
+alphabetSequenceGreekLettersCap = ['Alpha', 'Beta', 'Gamma', 'Delta', 'Epsilon', 'Zeta', 'Eta', 'Theta', 'Iota',
+                                   'Kappa', 'Lambda', 'Mu', 'Nu', 'Xi', 'Omicron', 'Pi', 'Rho', 'Sigma', 'Tau',
+                                   'Upsilon', 'Phi', 'Chi', 'Psi', 'Omega']
 
 
 def string_to_float(potential_float):
@@ -85,15 +99,13 @@ def list_to_string(subject_list, sep=''):
     Converts a list to a string.
 
     Parameters
-    ----------
-    subject_list : list
-        List to be converted to a string.
-    sep : str, optional
-        Delimiter in between list elements in the string. The default value is ''.
+        subject_list : list
+            The list that is to be converted to a string.
+        sep : str, optional
+            Delimiter in between list elements in the string. The default value is ''.
 
     Returns
-    -------
-    String from the list elements with the set delimiter in between.
+        String from the list elements with the set delimiter in between.
 
     """
     fixed_list = [str(i) if not isinstance(i, str) else i for i in subject_list]  # fix non-str elements to str type
@@ -322,7 +334,6 @@ def replace(elems, reps, string, exclusions=None):
         # define updated string, with the exclusions temporarily removed
         excl_string = [[i] + [e] if i not in [k[0] for k in fixed_index_excl_list] else [i] + [''] for i, e in
                        indexer(string)]
-        print(excl_string)
     else:
         excl_string = indexer(string)
 
@@ -380,3 +391,123 @@ def replace(elems, reps, string, exclusions=None):
     else:
         res = list_to_string(temp_string)  # define result and convert to string
     return res
+
+
+def string_sorter(*lists, stype='size', reverse=False, otype='list'):
+    """
+    Sorts any amount of given lists of strings, according to the first list given, depending on the sorting type.
+
+    Parameters
+        *lists : list
+            The lists of strings in need of being sorted. Sorts all lists according to the strings in the first list.
+        stype : str, optional
+            Determines which sorting type should be used. If 'size', sorts after size of the string (in order of
+            smallest to largest). If 'alphabetic', sorts strings alphabetically. The default is 'size'.
+        reverse : bool, optional
+            Reverses the sorting order. The default is False.
+        otype : str, optional
+            Determines the output type. If 'list', a list of lists is created. If 'tuple' a tuple of tuples is created.
+            The default is 'list'.
+
+    Returns
+        A list/tuple of lists/tuples with the sorted strings. The output list/tuple sequence matches input.
+    """
+
+    # if sorting type is size, construct a uniform list including a corresponding size-list as first list
+    if stype == 'size':
+        uniform_list = list(zip([len(i) for i in lists[0]], *lists))
+        sorted_lists_pre = sorted(uniform_list, key=itemgetter(0), reverse=reverse)  # sort lists
+        sorted_lists = [i[1:] for i in sorted_lists_pre]  # remove size-list
+
+    # if sorting type is alphabetic, construct a uniform list from the given lists and conduct sorting
+    elif stype in ('alpha', 'alphabetic', 'alphabet'):
+        uniform_list = list(zip(*lists))
+        sorted_lists = sorted(uniform_list, key=itemgetter(0), reverse=reverse)  # sort lists
+    else:  # raise error if sorting type is unknown
+        raise ValueError(f'Sorting type \'{stype}\' is undefined.')
+
+    # define the output according to the output type
+    if otype == 'tuple':
+        output_lists = tuple(zip(*sorted_lists))
+    elif otype == 'list':
+        output_lists = [list(i) for i in list(zip(*sorted_lists))]
+    else:
+        raise ValueError(f'Output type \'{otype}\' is undefined.')
+
+    if len(output_lists) == 1:
+        output_lists = output_lists[0]
+
+    return output_lists
+
+
+def split(string, delim, no_blanks=True):
+    """
+    Function that splits around a given delimiter, rather than at the given delimiter as the python .split() function.
+
+    Parameters
+        string : str
+            The string in which the splitting should be made in.
+        delim : str
+            Identifier for which a split should be made around.
+        no_blanks : bool, optional
+            In some cases, blank list elements like ['', ''] may be created, if this parameter is True, those blanks
+            are removed. The default is True.
+
+    Returns
+        A list containing the parts from the split.
+    """
+    temp_list = string.replace(delim, '𒐫𐩕𒐫').split('𒐫')
+    pre_string = [delim if i == '𐩕' else i for i in temp_list]
+
+    if no_blanks:
+        try:
+            result_string = [i for i in pre_string if i != '']
+        except ValueError:
+            result_string = pre_string
+    else:
+        result_string = pre_string
+
+    return result_string
+
+
+def multi_split(string, items, no_blanks=True):
+    """
+    An advanced version of the util.split() function. Splits the given string around every given item, and yields a
+    list with the result. Splits around items in order of occurrence, thus, no items that have already been iterated
+    through can be split again.
+
+    Parameters
+        string : str
+            The string to split in.
+        items : tuple
+            String elements that the function should split the main string around. This is a tuple of string elements.
+        no_blanks : bool, optional
+            In some cases, blank list elements can appear, this removes those elements.
+
+    Returns
+        List of the separated string parts.
+    """
+    # set initial values for iteration
+    temp_itr_str = [string]  # input string must be a list
+    iterated_items = []
+    remain_items = nest_checker(items, 'list')  # if items are not packed, pack them
+    string_wo_item = string  # define checker string, to avoid iterations, if item is not in the string
+
+    # while there are still items remaining, keep iterating
+    while remain_items:
+        item = remain_items[0]  # define the current item for checking as the first of the remaining
+
+        # if an item string is found in the main string, split the string around the item IF the item
+        #   has not already been iterated through
+        if item in string_wo_item:
+            temp_itr_str = [[i] if i in iterated_items else split(i, item, no_blanks=no_blanks) for
+                            i in temp_itr_str]
+            temp_itr_str = list(chain.from_iterable(temp_itr_str))
+
+        # update temporary lists
+        iterated_items += [item]
+        remain_items.remove(item)  # remove the current item from the remaining items
+        string_wo_item = list_to_string(string_wo_item.split(item))  # update checker string
+
+    # return result list
+    return temp_itr_str
